/*
 * BatteryPath.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.List;

/**
 * Represents a battery path as a list of actions leading to a given battery.
 *
 * Note: Natural ordering of this class is incosistent with equals!
 *
 * @author yanas
 */
final class BatteryPath extends Path {

    /** number of points that might be gained by recharge in the last game round */
    private final int pointsGainedByRecharge;
    /** precalculated profitability factor */
    private final float profitabilityFactor;

    /**
     * Constructs a non-existing BatteryPath object.
     */
    public BatteryPath() {
	// call parent constructor
	super();

	// initialize remaining attributes
	pointsGainedByRecharge = 0;
	profitabilityFactor = 0.0f;
    }

    /**
     * Constructs a BatteryPath object using the given parameters.
     * 
     * @param goal position of the goal in the world
     * @param actions a list of actions leading to the goal specified
     * @param remainingRoundsCount number of remaining game rounds
     * @param rechargeRobotEnergy robot energy gained by recharge
     */
    public BatteryPath(Position goal, List<BotAction> actions, int remainingRoundsCount,
	    int rechargeRobotEnergy) {
	// call parent constructor
	super(goal, actions);

	// initialize remaining attributes
	pointsGainedByRecharge = rechargeRobotEnergy;
	profitabilityFactor = (float) pointsGainedByRecharge
		/ ((remainingRoundsCount + 1) * remainingRoundsCount / 2 + length());
    }

    /**
     * Returns the number of points gained by recharge.
     *
     * @return the number of points gained by recharge
     */
    public int getPointsGainedByRecharge() {
	return pointsGainedByRecharge;
    }

    /**
     * Returns the profitability factor based on points gained by recharge and path length.
     *
     * @return the profitability factor of this treasure path
     */
    public float getProfitabilityFactor() {
	return profitabilityFactor;
    }

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (super.toString() + "; " + getPointsGainedByRecharge() + "; "
		+ getProfitabilityFactor());
    }

}
