/*
 * BatteryPathFinder.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;

/**
 * Represents an battery path finder based on A* algorithm.
 * 
 * This implementations relies on an instance of the WorldWrapper class which should be
 * provided during construction of the ShootingPathFinder. The instance of the WorldWrapper
 * class represents a world in which the path finding is carried out. It also contains the
 * robot for which the path finding should be done.
 *
 * BatteryPathFinder does care about resting. It finds the shortest path (a sequence of
 * actions) between the robot's current position and the goal position. The returned path
 * will probably be a bit longer than the one found by the PathFinder.
 * 
 * @author yanas
 *
 * @see TreasurePathFinder
 * @see ShootingPathFinder
 * @see PathFinder
 *
 * @see BatteryPath
 */
final class BatteryPathFinder {

    /**
     * Represents a search node for path finding.
     *
     * These nodes are used to build a search tree. Each node has its predecessor and therefore
     * you can climb the tree back up to the root node. This is useful when building a list of
     * actions from the start node to the goal node. Other attributes are used for heuristics.
     *
     * Note: Natural ordering of this class is incosistent with equals!
     *
     * @author  yanas
     */
    private static final class SearchNode implements Comparable<SearchNode> {

	/** this search node's predecessor */
	private final SearchNode predecessor;
	/** current oriented position */
	private final OrientedPosition orientedPosition;
	/** actions required to get from predecessor's oriented position to the current one */
	private final BotAction[] actions;
	/** total number of actions from start oriented position to the current one */
	private final int actionsCost;
	/** estimated number of actions to the goal position */
	private final int estimatedCost;
	/** robot's remaining energy */
	private final int robotEnergy;

	/**
	 * Constructs a starting SearchNode object. This constructor should be called only once
	 * at the beginning of the path finding. Then you should use the second one as this
	 * constructor actually creates a root node.
	 *
	 * @param orientedPosition starting oriented position
	 * @param goalCost estimated number of actions to the goal position
	 * @param robotEnergy robot's remaining energy
	 */
	public SearchNode(OrientedPosition orientedPosition, int goalCost, int robotEnergy) {
	    // initialize all the attributes
	    this.predecessor = null;
	    this.orientedPosition = orientedPosition;
	    this.actions = null;
	    this.actionsCost = 0;
	    this.estimatedCost = goalCost + ((goalCost >= robotEnergy) ? 1 : 0);
	    this.robotEnergy = robotEnergy;
	}

	/**
	 * Constructs a child SearchNode object. This constructor should be called when a root
	 * node has already been created. Use this constructor when expanding a node.
	 *
	 * @param predecessor direct predecessor of this node
	 * @param orientedPosition current oriented position
	 * @param actions actions required to get from predecessor's oriented position
	 * to the current one
	 * @param goalCost estimated number of actions to the goal position
	 * @param robotEnergy robot's remaining energy
	 */
	public SearchNode(SearchNode predecessor, OrientedPosition orientedPosition,
		BotAction[] actions, int goalCost, int robotEnergy) {
	    // initialize all the attributes
	    this.predecessor = predecessor;
	    this.orientedPosition = orientedPosition;
	    this.actions = actions;
	    this.actionsCost = predecessor.getActionsCost() + actions.length;
	    this.estimatedCost = actionsCost + goalCost + ((goalCost >= robotEnergy) ? 1 : 0);
	    this.robotEnergy = robotEnergy;
	}

	/**
	 * Returns the predecessor of this node.
	 *
	 * @return predecessor of this node
	 */
	public SearchNode getPredecessor() {
	    return predecessor;
	}

	/**
	 * Returns the oriented position.
	 *
	 * @return the oriented position
	 */
	public OrientedPosition getOrientedPosition() {
	    return orientedPosition;
	}

	/**
	 * Returns the position.
	 *
	 * @return the position
	 */
	public Position getPosition() {
	    return orientedPosition.getPosition();
	}

	/**
	 * Returns the direction.
	 *
	 * @return the direction
	 */
	public Direction getDirection() {
	    return orientedPosition.getDirection();
	}

	/**
	 * Returns the array of actions.
	 *
	 * @return the array of actions
	 */
	public BotAction[] getActions() {
	    return actions;
	}

	/**
	 * Returns the actions cost needed to "get" from the root node to this one
	 *
	 * @return actions cost accumulated so far
	 */
	public int getActionsCost() {
	    return actionsCost;
	}

	/**
	 * Returns the estimated cost which doesn't have to be precise.
	 *
	 * @return estimated actions cost
	 */
	public int getEstimatedCost() {
	    return estimatedCost;
	}

	/**
	 * Returns the robot's remaining energy.
	 *
	 * @return robot's remaining energy
	 */
	public int getRobotEnergy() {
	    return robotEnergy;
	}

	/**
	 * Compares this instance with another search node for order.
	 *
	 * @param searchNode another search node to be compared with this one
	 *
	 * @return an integer representing the relation between the nodes
	 */
	public int compareTo(SearchNode searchNode) {
	    return (getEstimatedCost() - searchNode.getEstimatedCost());
	}

	/**
	 * Returns string representation of this instance.
	 *
	 * @return important data converted to a string
	 */
	public String toString() {
	    return (getOrientedPosition() + "; " + getActionsCost() + "; "
		    + getEstimatedCost() + "; " + getRobotEnergy());
	}

    }

    /** actions needed to move forward */
    private static final BotAction[] ACTIONS_MOVE_FORWARD = {BotAction.STEP};
    /** actions needed to move left */
    private static final BotAction[] ACTIONS_MOVE_LEFT = {BotAction.LEFT, BotAction.STEP};
    /** actions needed to move right */
    private static final BotAction[] ACTIONS_MOVE_RIGHT = {BotAction.RIGHT, BotAction.STEP};
    /** actions needed to move backward */
    private static final BotAction[] ACTIONS_MOVE_BACKWARD = {BotAction.RIGHT,
	    BotAction.RIGHT, BotAction.STEP};

    /** a world in which the search is carried out */
    private final WorldWrapper world;

    /** a priority queue of unexplored search nodes ordered by estimated cost */
    private final Queue<SearchNode> searchNodes = new PriorityQueue<SearchNode>();
    /** array of lowest actions costs per position */
    private final int actionsCosts[][];

    /**
     * Constructs an BatteryPathFinder object. An instance of the WorldWrapper class passed
     * to this constructor is used every time you want to find a path to a goal position and
     * cannot be modified.
     * 
     * @param world a world in which the search is carried out
     */
    public BatteryPathFinder(WorldWrapper world) {
	// save the world
	this.world = world;

	// create and initialize the actions cost array based on the world's dimensions
	actionsCosts = new int[world.getDimensionX()][world.getDimensionY()];
	clearActionsCosts();
    }

    /**
     * Finds the best battery path based on the current world state.
     *
     * @return a BatteryPath object representing best battery path
     */
    public BatteryPath findBestPath() {
	// prepare a list for battery paths
	List<BatteryPath> batteryPaths = new ArrayList<BatteryPath>();

	// for each battery in the world
	for (Position battery : world.getBatteries()) {
	    // find a battery path to the given battery
	    BatteryPath batteryPath = findPath(world.getNullBot(), battery);

	    // if the path exists
	    if (batteryPath.exists()) {
		// add it to the list
		batteryPaths.add(batteryPath);
	    }
	}

	// return a non-existing path if no existing path has been found
	if (batteryPaths.isEmpty()) {
	    return new BatteryPath();
	}

	// finally sort the list with battery paths
	Collections.sort(batteryPaths);

	// for each battery path found
	for (BatteryPath batteryPath : batteryPaths) {
	    // if the battery path doesn't exist
	    if (!batteryPath.exists()) {
		// break the loop
		break;
	    }

	    // prepare flag for the current batteryPath
	    boolean isBestBatteryPath = true;

	    // for each opponent robot
	    for (RobotWrapper opponentRobot : world.getOpponentRobots()) {
		// find opponents battery path to our goal
		BatteryPath opponentsPath = findPath(opponentRobot, batteryPath.getGoal());

		// if our battery path is longer than the opponent's battery path
		if (batteryPath.isLongerThan(opponentsPath)) {
		    // set the flag to false and break the loop
		    isBestBatteryPath = false;
		    break;
		}
	    }

	    // return the battery path if it is the best battery path
	    if (isBestBatteryPath) {
		return batteryPath;
	    }
	}

	return batteryPaths.get(0);
    }

    /**
     * Returns string representation of this instance. For debugging purposes.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (world + "; " + searchNodes);
    }

    /**
     * Finds the shortest path between the associated robot's oriented position and the goal
     * position. Resting and recharging is taken into consideration when finding a path. If
     * there is not enough game rounds to reach the goal, the returned path won't exist.
     *
     * @param robot a robot for which the finding is done
     * @param goal position to be reached
     *
     * @return an instance of the BatteryPath class representing the result path
     */
    private BatteryPath findPath(RobotWrapper robot, Position goal) {
	// calculate manhattan distance from start to goal
	int manhattanDistance = Position.manhattanDistance(robot.getPosition(), goal);

	// return non-existing path if the goal is unreachable
	if (world.containsObstacle(goal) || manhattanDistance >= world.remainingRoundsCount()) {
	    return new BatteryPath();
	}

	// create a starting search node
	SearchNode startNode = new SearchNode(robot.getOrientedPosition(), manhattanDistance,
		robot.getEnergy());

	// add the search node to the search nodes queue, set actions cost to zero
	searchNodes.add(startNode);
	setActionsCost(robot.getPosition(), 0);

	// expand the start search node by moving backward
	expandSearchNode(startNode, startNode.getOrientedPosition().moveBackward(),
		ACTIONS_MOVE_BACKWARD, goal);

	// while there are any search nodes in the search nodes queue
	while (!searchNodes.isEmpty()) {
	    // remove the best node from the nodes queue
	    SearchNode searchNode = searchNodes.remove();

	    // if the actions cost is higher or equal than the remaining rounds count
	    if (searchNode.getActionsCost() >= world.remainingRoundsCount()) {
		// break the loop to finish searching
		break;
	    }

	    // if the goal has been reached
	    if (goal.equals(searchNode.getPosition())) {
		// create an actions list
		LinkedList<BotAction> actions = new LinkedList<BotAction>();

		// while the node is not null and the node's actions are not null
		while (searchNode != null && searchNode.getActions() != null) {
		    // loop through all the actions in the array
		    for (int i = 0; i < searchNode.getActions().length; i++) {
			// add the action at the beginning of the actions list
			actions.add(i, searchNode.getActions()[i]);
		    }

		    // save this node's predecessor
		    searchNode = searchNode.getPredecessor();
		}

		// clear both nodes queue and actions costs
		searchNodes.clear();
		clearActionsCosts();

		return new BatteryPath(goal, actions, world.remainingRoundsCount(),
			world.getMaxRobotEnergy());
	    }

	    // expand the current search node by moving forward
	    expandSearchNode(searchNode, searchNode.getOrientedPosition().moveForward(),
		    ACTIONS_MOVE_FORWARD, goal);

	    // expand the current search node by moving to left
	    expandSearchNode(searchNode, searchNode.getOrientedPosition().moveLeft(),
		    ACTIONS_MOVE_LEFT, goal);

	    // expand the current search node by moving to right
	    expandSearchNode(searchNode, searchNode.getOrientedPosition().moveRight(),
		    ACTIONS_MOVE_RIGHT, goal);
	}

	// clear both nodes queue and actions costs
	searchNodes.clear();
	clearActionsCosts();

	return new BatteryPath();
    }

    /**
     * Expands the given search node using given oriented position, actions and goal.
     *
     * @param searchNode a search node to be expanded
     * @param orientedPosition oriented position we want to move to from the current position
     * @param actions an array of actions required by the above movement
     * @param goal a position we want to reach
     */
    private void expandSearchNode(SearchNode searchNode, OrientedPosition orientedPosition,
	    BotAction[] actions, Position goal) {
	// save the position we will move to by expanding the given seach node
	Position position = orientedPosition.getPosition();

	// if there is no obstacle at the new position
	if (!world.containsObstacle(position)) {
	    // prepare robot energy which will be used when no resting is needed
	    int robotEnergy = searchNode.getRobotEnergy() - actions.length;

	    // if the robot didn't have enough energy to carry out all the actions
	    if (robotEnergy < 1) {
		// get the robot energy
		robotEnergy = searchNode.getRobotEnergy();

		// create an array for modified actions, set rest actions count to zero
		BotAction[] modifiedActions = new BotAction[2 * actions.length];
		int restActionsCount = 0;

		// loop through all the actions
		for (int i = 0; i < actions.length; i++) {
		    // if energy equals to one
		    if (robotEnergy == 1) {
			// save the rest action, set the robot energy to rest energy
			modifiedActions[i + restActionsCount++] = BotAction.REST;
			robotEnergy = world.getRestRobotEnergy();
		    }

		    // save the action and decrement the robot energy
		    modifiedActions[i + restActionsCount] = actions[i];
		    robotEnergy--;
		}

		// create a new array of actions of the correct size
		actions = new BotAction[actions.length + restActionsCount];

		// loop through all the array
		for (int i = 0; i < actions.length; i++) {
		    // save the modified action
		    actions[i] = modifiedActions[i];
		}
	    }

	    // calculate the total actions cost from start to this position
	    int totalActionsCost = searchNode.getActionsCost() + actions.length;

	    // if the actions cost is lower than (or equal to) the best one stored so far
	    if (totalActionsCost <= getActionsCost(position)) {
		// create and add a new node to the search nodes queue
		searchNodes.add(new SearchNode(searchNode, orientedPosition, actions,
			Position.manhattanDistance(position, goal), robotEnergy));

		// set new actions cost for this position calculated earlier
		setActionsCost(position, totalActionsCost);
	    }
	}
    }

    /**
     * Returns the action cost for the given position.
     *
     * @param position a position we want to know about
     *
     * @return a non-zero integer when the position has already been visited, integer max
     * otherwise
     */
    private int getActionsCost(Position position) {
	return actionsCosts[position.getX()][position.getY()];
    }

    /**
     * Sets action cost for the given position.
     *
     * @param position a position for which the actions is to be set
     * @param actionsCost new actions cost to be set
     */
    private void setActionsCost(Position position, int actionsCost) {
	actionsCosts[position.getX()][position.getY()] = actionsCost;
    }

    /**
     * Clears the actions costs array by filling each cell with int max value. This value
     * means that a position has not yet been visited.
     */
    private void clearActionsCosts() {
	// loop through all the arrays in the actions costs array
	for (int i = 0; i < world.getDimensionX(); i++) {
	    // fill each array with max int value as the highest actions cost
	    Arrays.fill(actionsCosts[i], Short.MAX_VALUE);
	}
    }

}
