/*
 * NullBot.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.Random;

/**
 * Represents our NullBot's strategy.
 */
public final class NullBot implements Strategy {

    /** a world set during initialization */
    private World worldSet;
    /** a robot set during initialization */
    private Robot robotSet;

    /** a world wrapper around the world set */
    private WorldWrapper world;
    /** a robot wrapper around our robot */
    private RobotWrapper nullbot;

    /** a treasure path finder */
    private TreasurePathFinder treasurePathFinder;
    /** a shooting path finder */
    private ShootingPathFinder shootingPathFinder;
    /** a battery path finder */
    private BatteryPathFinder batteryPathFinder;

    /** a number of consecutive colliding actions */
    private int collidingActionsCount;
    /** a number of actions after which obstacle avoidance is requested */
    private int collisionAvoidanceTreshold;

    /** a random numbers generator */
    private Random random = new Random();

    /**
     * Sets the world in which the game will be played.
     *
     * @param world a world to be used
     */
    public void setWorld(World world) {
	worldSet = world;
    }

    /**
     * Sets the robot for our team.
     *
     * @param robot a robot that belongs to our team
     */
    public void setRobot(Robot robot) {
        robotSet = robot;
    }

    /**
     * Performs necessary initialization before the game starts.
     */
    public void initialize() {
	// initialize wrappers
	nullbot = new RobotWrapper(robotSet);
	world = new WorldWrapper(worldSet, nullbot);

	// initialize all the path finders
	treasurePathFinder = new TreasurePathFinder(world);
	shootingPathFinder = new ShootingPathFinder(world);
	batteryPathFinder = new BatteryPathFinder(world);

	// set colliding actions count and avoidance treshold to zero
	collidingActionsCount = 0;
	collisionAvoidanceTreshold = 0;
    }

    /**
     * The core method of entire NullBot strategy which performs all the decisions using
     * wrappers and path finders created during initialization of the game.
     *
     * @return a BotAction we think is the best choice
     */
    public BotAction decide() {
	// if we have enough energy to make a step
	if (nullbot.getEnergy() > 1) {
	    // increment the colliding actions count
	    collidingActionsCount++;

	    // if a collision cannot occur in the next round
	    if (!world.mayCollisionOccur()) {
		// reset the colliding actions count, set new collision avoidance treshold
		collidingActionsCount = 0;
		collisionAvoidanceTreshold = 2 + random.nextInt(3);
	    }
	}

	// find the best treasure, shooting and battery paths
	TreasurePath treasurePath = treasurePathFinder.findBestPath(
		collidingActionsCount > collisionAvoidanceTreshold);
	ShootingPath shootingPath = shootingPathFinder.findBestPath(true);
	BatteryPath batteryPath = batteryPathFinder.findBestPath();

	// if the treasure path exists and more profitable than the other paths
	if (treasurePath.exists() && treasurePath.isMoreProfitableThan(shootingPath)
		&& treasurePath.isMoreProfitableThan(batteryPath)) {
	    // follow the path if it's not empty
	    if (!treasurePath.isEmpty()) {
		return performAction(treasurePath.follow(), nullbot.getEnergy() - 1);
	    }

	    return performAction(BotAction.TAKE, treasurePath.getTreasureValue()
		    + nullbot.getEnergy() - 1);
	}

	// if the shooting path exists and is more profitable than the battery path
	if (shootingPath.exists() && shootingPath.isMoreProfitableThan(batteryPath)) {
	    // follow the path if it's not empty
	    if (!shootingPath.isEmpty()) {
		return performAction(shootingPath.follow(), nullbot.getEnergy() - 1);
	    }

	    return performAction(BotAction.SHOOT, shootingPath.getPointsGainedByShooting()
		    + nullbot.getEnergy() - 1);
	}

	// if the battery path exists
	if (batteryPath.exists()) {
	    // follow the path if it's not empty
	    if (!batteryPath.isEmpty()) {
		return performAction(batteryPath.follow(), nullbot.getEnergy() - 1);
	    }
	    else if (world.isLastRound()) {
		return performAction(BotAction.RECHARGE,batteryPath.getPointsGainedByRecharge());
	    }
	}

	// find the best shooting path without any moving around
	shootingPath = shootingPathFinder.findBestPath(false);

	// if such a shooting path exists
	if (shootingPath.exists()) {
	    // follow the path if it's not empty
	    if (!shootingPath.isEmpty()) {
		return performAction(shootingPath.follow(), nullbot.getEnergy() - 1);
	    }

	    return performAction(BotAction.SHOOT, shootingPath.getPointsGainedByShooting()
		    + nullbot.getEnergy() - 1);
	}

	// turn randomly if anyone may shoot at us
	if (shootingPathFinder.mayAnyoneShootAtUs()) {
	    return performAction((((random.nextInt() % 2) == 0) ? BotAction.LEFT
		    : BotAction.RIGHT), nullbot.getEnergy() - 1);
	}

	return BotAction.REST;
    }

    /**
     * Returns the given action if the robot has enough energy or if its performing in the last
     * game round is more profitable than resting, otherwise it returns the rest action.
     *
     * @param action an action to be taken into consideration
     * @param pointsGained a number of points that'll be gained by performing by the above action
     *
     * @return the best action to be made in the current game round
     */
    private BotAction performAction(BotAction action, int pointsGained) {
	// rest if the robot has little energy or the action is not as profitable as resting
	if ((nullbot.getEnergy() == 1 && !world.isLastRound() && action != BotAction.RECHARGE)
		|| (world.isLastRound() && pointsGained < (nullbot.getEnergy()
		    + world.getRestRobotEnergy() - 1))) {
	    return BotAction.REST;
	}

	return action;
    }

}
