/*
 * OrientedPosition.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Robot.Direction;

/**
 * Represents an immutable and oriented 2D position in the world.
 *
 * @author yanas
 */
final class OrientedPosition {

    /** position */
    private final Position position;
    /** direction */
    private final Direction direction;

    /**
     * Constructs an OrientedPosition object using given position and direction.
     *
     * @param position position to be set
     * @param direction direction to be set
     */
    public OrientedPosition(Position position, Direction direction) {
	// save both position and direction
	this.position = position;
	this.direction = direction;
    }

    /**
     * Returns the position.
     *
     * @return stored instance of Position class
     */
    public Position getPosition() {
	return position;
    }

    /**
     * Returns the direction.
     *
     * @return stored instance of Direction class
     */
    public Direction getDirection() {
	return direction;
    }

    /**
     * Returns this oriented position turned left.
     *
     * @return a new instance of OrientedPosition class turned left
     */
    public OrientedPosition turnLeft() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition(), Direction.LEFT);
	    case RIGHT:
		return new OrientedPosition(getPosition(), Direction.UP);
	    case DOWN:
		return new OrientedPosition(getPosition(), Direction.RIGHT);
	    case LEFT:
		return new OrientedPosition(getPosition(), Direction.DOWN);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns this oriented position turned right.
     *
     * @return a new instance of OrientedPosition class turned right
     */
    public OrientedPosition turnRight() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition(), Direction.RIGHT);
	    case RIGHT:
		return new OrientedPosition(getPosition(), Direction.DOWN);
	    case DOWN:
		return new OrientedPosition(getPosition(), Direction.LEFT);
	    case LEFT:
		return new OrientedPosition(getPosition(), Direction.UP);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns this oriented position turned around.
     *
     * @return a new instance of OrientedPosition class turned around
     */
    public OrientedPosition turnAround() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition(), Direction.DOWN);
	    case RIGHT:
		return new OrientedPosition(getPosition(), Direction.LEFT);
	    case DOWN:
		return new OrientedPosition(getPosition(), Direction.UP);
	    case LEFT:
		return new OrientedPosition(getPosition(), Direction.RIGHT);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns this oriented position moved forward.
     *
     * @return a new instance of OrientedPosition class moved forward
     */
    public OrientedPosition moveForward() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition().offset(0, -1), Direction.UP);
	    case RIGHT:
		return new OrientedPosition(getPosition().offset(1, 0), Direction.RIGHT);
	    case DOWN:
		return new OrientedPosition(getPosition().offset(0, 1), Direction.DOWN);
	    case LEFT:
		return new OrientedPosition(getPosition().offset(-1, 0), Direction.LEFT);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns this oriented position moved left.
     *
     * @return a new instance of OrientedPosition class moved left
     */
    public OrientedPosition moveLeft() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition().offset(-1, 0), Direction.LEFT);
	    case RIGHT:
		return new OrientedPosition(getPosition().offset(0, -1), Direction.UP);
	    case DOWN:
		return new OrientedPosition(getPosition().offset(1, 0), Direction.RIGHT);
	    case LEFT:
		return new OrientedPosition(getPosition().offset(0, 1), Direction.DOWN);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns this oriented position moved right.
     *
     * @return a new instance of OrientedPosition class moved right
     */
    public OrientedPosition moveRight() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition().offset(1, 0), Direction.RIGHT);
	    case RIGHT:
		return new OrientedPosition(getPosition().offset(0, 1), Direction.DOWN);
	    case DOWN:
		return new OrientedPosition(getPosition().offset(-1, 0), Direction.LEFT);
	    case LEFT:
		return new OrientedPosition(getPosition().offset(0, -1), Direction.UP);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns this oriented position moved backward.
     *
     * @return a new instance of OrientedPosition class moved backward
     */
    public OrientedPosition moveBackward() {
	// which direction are we looking now?
	switch (getDirection()) {
	    // handle correct values
	    case UP:
		return new OrientedPosition(getPosition().offset(0, 1), Direction.DOWN);
	    case RIGHT:
		return new OrientedPosition(getPosition().offset(-1, 0), Direction.LEFT);
	    case DOWN:
		return new OrientedPosition(getPosition().offset(0, -1), Direction.UP);
	    case LEFT:
		return new OrientedPosition(getPosition().offset(1, 0), Direction.RIGHT);

	    // this should never happen
	    default:
		return null;
	}
    }

    /**
     * Returns true if an object is equal to this instance, false otherwise.
     *
     * @param object an object to be compared with this instance
     *
     * @return true if the objects are equal
     */
    public boolean equals(Object object) {
	// return true if both instances are equal
	if (object == this) {
	    return true;
	}

	// return false if object is of different type
	if (!(object instanceof OrientedPosition)) {
	    return false;
	}

	// save the object as an instance of the class OrientedPosition
	OrientedPosition orientedPosition = (OrientedPosition) object;

	return (getPosition().equals(orientedPosition.getPosition())
		&& getDirection().equals(orientedPosition.getDirection()));
    }

    /**
     * Returns hash code of this instance.
     *
     * @return hash code
     */
    public int hashCode() {
	int result = 17;
	result = 37 * result + getPosition().hashCode();
	result = 37 * result + getDirection().hashCode();

	return result;
    } 

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (getPosition() + " " + getDirection());
    }

}
