/*
 * Path.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.Collections;
import java.util.List;

/**
 * Represents a path as a list of actions leading to a given goal. Each instance of this Path
 * class is meant to be used only in one round. Therefore only one action can be retrieved.
 *
 * Note: Natural ordering of this class is incosistent with equals!
 *
 * @author yanas
 */
abstract class Path implements Comparable<Path> {

    /** goal of this path */
    private final Position goal;
    /** a list of actions leading to the goal */
    private final List<BotAction> actions;

    /**
     * Constructs a non-existing Path object.
     */
    public Path() {
	goal = null;
	actions = null;
    }

    /**
     * Constructs a Path object using the given parameters.
     *
     * @param goal position of the goal in the world
     * @param actions a list of actions leading to the goal specified
     */
    public Path(Position goal, List<BotAction> actions) {
	this.goal = goal;
	this.actions = actions;
    }

    /**
     * Returns the goal.
     *
     * @return the goal
     */
    public Position getGoal() {
	return goal;
    }

    /**
     * Returns the list of actions.
     *
     * @return the list of actions
     */
    public List<BotAction> getActions() {
	return (exists() ? Collections.unmodifiableList(actions) : null);
    }

    /**
     * Returns true if the path to the goal exists, false otherwise.
     *
     * @return true if the path exists
     */
    public boolean exists() {
	return (actions != null);
    }

    /**
     * Returns true if the path is empty, false otherwise.
     *
     * @return true if the path is empty
     */
    public boolean isEmpty() {
	return actions.isEmpty();
    }

    /**
     * Returns the length of the path.
     *
     * @return the actual length if the path exists, short max otherwise
     */
    public int length() {
	return (exists() ? actions.size() : Short.MAX_VALUE);
    }

    /**
     * Returns true if this path is shorter than the second one.
     *
     * @param path another path to be compared with this one
     *
     * @return true if this path is shorter
     */
    public boolean isShorterThan(Path path) {
	return (length() < path.length());
    }

    /**
     * Returns true if this path is longer than the second one.
     *
     * @param path another path to be compared with this one
     *
     * @return true if this path is longer
     */
    public boolean isLongerThan(Path path) {
	return (length() > path.length());
    }

    /**
     * Returns the profitability factor of this path which is implementation dependent.
     *
     * @return the profitability factor of this path
     */
    abstract public float getProfitabilityFactor();

    /**
     * Returns true if this path is more profitable than the second one.
     *
     * @param path another path to be compared with this one
     *
     * @return true if this path is more profitable
     */
    public boolean isMoreProfitableThan(Path path) {
	return (getProfitabilityFactor() > path.getProfitabilityFactor());
    }

    /**
     * Returns true if this path is less profitable than the second one.
     *
     * @param path another path to be compared with this one
     *
     * @return true if this path is less profitable
     */
    public boolean isLessProfitableThan(Path path) {
	return (getProfitabilityFactor() < path.getProfitabilityFactor());
    }

    /**
     * Follows the path by returning the first action from the list.
     *
     * @return BotAction that is first in the path
     */
    public BotAction follow() {
	return actions.get(0);
    }

    /**
     * Compares this instance with another path for order.
     *
     * @param path a path to be compared with this one
     *
     * @return an integer representing the relation between the paths
     */
    public int compareTo(Path path) {
	return Math.round(Math.signum(path.getProfitabilityFactor() - getProfitabilityFactor()));
    }

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (getGoal() + "; " + getActions());
    }

}
