/*
 * PathFinder.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Robot.Direction;

import java.util.Arrays;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Set;

/**
 * Represents a general path finder based on A* algorithm.
 * 
 * This implementations relies on an instance of the WorldWrapper class which should be
 * provided during construction of the PathFinder. The instance of the WorldWrapper class
 * represents a world in which the path finding is carried out.
 * 
 * PathFinder does not care about recharging or resting. It simply checks if there exists
 * a path (a sequence of actions) between start oriented position and the goal position.
 * 
 * @author yanas
 *
 * @see TreasurePathFinder
 * @see ShootingPathFinder
 */
final class PathFinder {

    /**
     * Represents a search node for path finding.
     *
     * These nodes are used to build a search tree. Although each node has its predecessor,
     * it is not stored in the search node as it's not necessary. All the attributes stored
     * are used for heuristics.
     *
     * Note: Natural ordering of this class is incosistent with equals!
     *
     * @author  yanas
     */
    private static final class SearchNode implements Comparable<SearchNode> {

	/** current oriented position */
	private final OrientedPosition orientedPosition;
	/** total number of actions from start oriented position to the current one */
	private final int actionsCost;
	/** estimated number of actions to the goal position */
	private final int estimatedCost;

	/**
	 * Constructs a starting SearchNode object. This constructor should be called only once
	 * at the beginning of the path finding. Then you should use the second one as this
	 * constructor actually creates a root node.
	 *
	 * @param orientedPosition starting oriented position
	 * @param goalCost estimated number of actions to the goal position
	 */
	public SearchNode(OrientedPosition orientedPosition, int goalCost) {
	    // initialize all the attributes
	    this.orientedPosition = orientedPosition;
	    this.actionsCost = 0;
	    this.estimatedCost = goalCost;
	}

	/**
	 * Constructs a child SearchNode object. This constructor should be called when a root
	 * node has already been created. Use this constructor when expanding a node.
	 *
	 * @param orientedPosition current oriented position
	 * @param actionsCost number of actions required to get from start node's oriented
	 * position to the current one
	 * @param goalCost estimated number of actions to the goal position
	 */
	public SearchNode(OrientedPosition orientedPosition, int actionsCost, int goalCost) {
	    // initialize all the attributes
	    this.orientedPosition = orientedPosition;
	    this.actionsCost = actionsCost;
	    this.estimatedCost = actionsCost + goalCost;
	}

	/**
	 * Returns the oriented position.
	 *
	 * @return the oriented position
	 */
	public OrientedPosition getOrientedPosition() {
	    return orientedPosition;
	}

	/**
	 * Returns the position.
	 *
	 * @return the position
	 */
	public Position getPosition() {
	    return orientedPosition.getPosition();
	}

	/**
	 * Returns the direction.
	 *
	 * @return the direction
	 */
	public Direction getDirection() {
	    return orientedPosition.getDirection();
	}

	/**
	 * Returns the actions cost needed to "get" from the root node to this one
	 *
	 * @return actions cost accumulated so far
	 */
	public int getActionsCost() {
	    return actionsCost;
	}

	/**
	 * Returns the estimated cost which doesn't have to be precise.
	 *
	 * @return estimated actions cost
	 */
	public int getEstimatedCost() {
	    return estimatedCost;
	}

	/**
	 * Compares this instance with another search node for order.
	 *
	 * @param searchNode another search node to be compared with this one
	 *
	 * @return an integer representing the relation between the nodes
	 */
	public int compareTo(SearchNode searchNode) {
	    return (getEstimatedCost() - searchNode.getEstimatedCost());
	}

	/**
	 * Returns string representation of this instance.
	 *
	 * @return important data converted to a string
	 */
	public String toString() {
	    return (getOrientedPosition() + "; " + getActionsCost() + "; " + getEstimatedCost());
	}

    }

    /** actions needed to move forward */
    private static final int ACTIONS_COST_MOVE_FORWARD = 1;
    /** actions needed to move left */
    private static final int ACTIONS_COST_MOVE_LEFT = 2;
    /** actions needed to move right */
    private static final int ACTIONS_COST_MOVE_RIGHT = 2;
    /** actions needed to move backward */
    private static final int ACTIONS_COST_MOVE_BACKWARD = 3;

    /** a world in which the search is carried out */
    private final WorldWrapper world;

    /** a priority queue of unexplored search nodes ordered by estimated cost */
    private final Queue<SearchNode> searchNodes = new PriorityQueue<SearchNode>();
    /** array of lowest actions costs per position */
    private final int actionsCosts[][];

    /**
     * Constructs a PathFinder object. Arguments passed to this constructor are used every
     * time you want to find a path to a goal position and cannot be modified.
     * 
     * 
     * @param world a world in which the search is carried out
     */
    public PathFinder(WorldWrapper world) {
	// save reference to the world
	this.world = world;

	// create and initialize the actions cost array based on the world's dimensions
	actionsCosts = new int[world.getDimensionX()][world.getDimensionY()];
	clearActionsCosts();
    }

    /**
     * Returns true if there exists a path between the start oriented position and the goal
     * position, false otherwise.
     *
     * @param start starting oriented position
     * @param goal position to be reached
     *
     * @return true if the path exists
     */
    public boolean pathExists(OrientedPosition start, Position goal) {
	// return -1 if the goal is not accessible
	if (world.containsWall(goal)) {
	    return false;
	}

	// create a starting search node
	SearchNode startNode = new SearchNode(start,
		Position.manhattanDistance(start.getPosition(), goal));

	// add the search node to the search nodes queue, set actions cost to zero
	searchNodes.add(startNode);
	setActionsCost(start.getPosition(), 0);

	// expand the start search node by moving backward
	createSearchNode(startNode.getOrientedPosition().moveBackward(),
		startNode.getActionsCost() + ACTIONS_COST_MOVE_BACKWARD, goal);

	// while there are any search nodes in the search nodes queue
	while (!searchNodes.isEmpty()) {
	    // remove the best node from the nodes queue
	    SearchNode searchNode = searchNodes.remove();

	    // if the goal has been reached
	    if (goal.equals(searchNode.getPosition())) {
		// clear both nodes queue and actions costs
		searchNodes.clear();
		clearActionsCosts();

		return true;
	    }

	    // expand the current search node by moving forward
	    createSearchNode(searchNode.getOrientedPosition().moveForward(),
		    searchNode.getActionsCost() + ACTIONS_COST_MOVE_FORWARD, goal);

	    // expand the current search node by moving to left
	    createSearchNode(searchNode.getOrientedPosition().moveLeft(),
		    searchNode.getActionsCost() + ACTIONS_COST_MOVE_LEFT, goal);

	    // expand the current search node by moving to right
	    createSearchNode(searchNode.getOrientedPosition().moveRight(),
		    searchNode.getActionsCost() + ACTIONS_COST_MOVE_RIGHT, goal);
	}

	// clear both nodes queue and actions costs
	searchNodes.clear();
	clearActionsCosts();

	return false;
    }

    /**
     * Returns true if there exists any path between the start position and each given goal
     * position, false otherwise.
     *
     * @param start starting oriented position
     * @param goals set of goals to be reached
     *
     * @return true if any path exists
     */
    public boolean anyPathExists(OrientedPosition start, Set<Position> goals) {
	// for each goal in the list
	for (Position goal : goals) {
	    // return true if the path exists
	    if (pathExists(start, goal)) {
		return true;
	    }
	}

	return false;
    }

    /**
     * Returns string representation of this instance. For debugging purposes.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (world + "; " + searchNodes);
    }

    /**
     * Creates a new search node using given oriented position, actions cost and goal.
     *
     * @param orientedPosition oriented position we want to move to from the current position
     * @param actionsCost number of actions consumed from the start position to the above one
     * @param goal a position we want to reach
     */
    private void createSearchNode(OrientedPosition orientedPosition, int actionsCost,
	    Position goal) {
	// save the position we will move to by expanding the given seach node
	Position position = orientedPosition.getPosition();

	// if there is no wall at the new position
	if (!world.containsWall(position)) {
	    // if the actions cost is lower than the best one stored so far
	    if (actionsCost < getActionsCost(position)) {
		// create and add a new node to the search nodes queue
		searchNodes.add(new SearchNode(orientedPosition, actionsCost,
			Position.manhattanDistance(position, goal)));

		// set new actions cost for this position calculated earlier
		setActionsCost(position, actionsCost);
	    }
	}
    }

    /**
     * Returns the action cost for the given position.
     *
     * @param position a position we want to know about
     *
     * @return a non-zero integer when the position has already been visited, integer max
     * otherwise
     */
    private int getActionsCost(Position position) {
	return actionsCosts[position.getX()][position.getY()];
    }

    /**
     * Sets action cost for the given position.
     *
     * @param position a position for which the actions is to be set
     * @param actionsCost new actions cost to be set
     */
    private void setActionsCost(Position position, int actionsCost) {
	actionsCosts[position.getX()][position.getY()] = actionsCost;
    }

    /**
     * Clears the actions costs array by filling each cell with int max value. This value
     * means that a position has not yet been visited.
     */
    private void clearActionsCosts() {
	// loop through all the arrays in the actions costs array
	for (int i = 0; i < world.getDimensionX(); i++) {
	    // fill each array with max int value as the highest actions cost
	    Arrays.fill(actionsCosts[i], Short.MAX_VALUE);
	}
    }

}
