/*
 * Position.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

/**
 * Represents an immutable 2D position in the world.
 *
 * @author yanas
 */
final class Position {

    /** the x-coordinate */
    private final int x;
    /** the y-coordinate */
    private final int y;

    /**
     * Constructs a Position object using given x and y coordinates.
     *
     * @param x the x-coordinate
     * @param y the y-coordinate
     */
    public Position(int x, int y) {
	this.x = x;
	this.y = y;
    }

    /**
     * Returns the x-coordinate.
     *
     * @return the x-coordinate
     */
    public int getX() {
	return x;
    }
    
    /**
     * Returns the y-coordinate.
     *
     * @return the y-coordinate
     */
    public int getY() {
	return y;
    }

    /**
     * Returns this position offset by the given offsets.
     *
     * @param dx the x offset
     * @param dy the y offset
     *
     * @return new instance of Position class offset by dx and dy
     */
    public Position offset(int dx, int dy) {
	return new Position(getX() + dx, getY() + dy);
    }

    /**
     * Calculates manhattan distance between start and goal positions.
     *
     * @param start start Position
     * @param goal goal Position
     *
     * @return manhattan distance between start and goal
     */
    public static int manhattanDistance(Position start, Position goal) {
	return (Math.abs(start.getX() - goal.getX()) + Math.abs(start.getY() - goal.getY()));
    }

    /**
     * Returns true if an object is equal to this instance, false otherwise.
     *
     * @param object an object to be compared with this instance
     *
     * @return true if the objects are equal
     */
    public boolean equals(Object object) {
	// return true if both instances are equal
	if (object == this) {
	    return true;
	}

	// return false if object is of different type
	if (!(object instanceof Position)) {
	    return false;
	}

	// save the object as an instance of the class Position
	Position position = (Position) object;

	return (getX() == position.getX() && getY() == position.getY());
    }

    /**
     * Returns hash code of this instance.
     *
     * @return hash code
     */
    public int hashCode() {
	int result = 17;
	result = 37 * result + getX();
	result = 37 * result + getY();

	return result;
    }

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return ("[" + getX() + ", " + getY() + "]");
    }

}
