/*
 * RobotWrapper.java
 *
 * @author network_gauner
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy.BotAction;

/**
 * A wrapper around the Robot class providing more functionality.
 *
 * @author network_gauner
 * @author yanas
 */
final class RobotWrapper {

    /** wrapped robot. */
    private Robot robot;

    /**
     * Constructs a RobotWrapper object around the given robot.
     *
     * @param robot a robot to be wrapped
     */
    public RobotWrapper(Robot robot) {
        this.robot = robot;
    }

    /**
     * Returns the wrapped robot.
     *
     * @return the wrapped robot
     */
    public Robot getWrappedRobot() {
        return this.robot;
    }

    /**
     * Returns the name of the robot.
     *
     * @return the name of the robot
     */
    public String getName() {
        return robot.getName();
    }

    /**
     * Returns the oriented position of the robot.
     *
     * @return the oriented position of the robot
     */
    public OrientedPosition getOrientedPosition() {
        return new OrientedPosition(getPosition(), getDirection());
    }

    /**
     * Returns the position of the robot.
     *
     * @return the position of the robot
     */
    public Position getPosition() {
        return new Position(robot.getPositionX(), robot.getPositionY());
    }

    /**
     * Returns the direction of the robot.
     *
     * @return the direction of the robot
     */
    public Direction getDirection() {
        return robot.getDirection();
    }

    /**
     * Returns the energy of the robot
     *
     * @return the energy of the robot
     */
    public int getEnergy() {
        return robot.getEnergy();
    }

    /**
     * Returns the number of points the robot has gained so far.
     *
     * @return the number of points 
     */
    public int getPoints() {
        return robot.getPoints();
    }

    /**
     * Returns the last action this robot made.
     *
     * @return the last action
     */
    public BotAction getLastAction() {
        return robot.getLastAction();
    }

    /**
     * Returns true if an object is equal to this instance, false otherwise.
     *
     * @param object an object to be compared with this instance
     *
     * @return true if the objects are equal
     */
    public boolean equals(Object object) {
	// return true if both instances are equal
	if (object == this) {
	    return true;
	}

	// return false if object is of different type
	if (!(object instanceof RobotWrapper)) {
	    return false;
	}

	// save the object as an instance of the class Position
	RobotWrapper robotWrapper = (RobotWrapper) object;

	return (getWrappedRobot().equals(robotWrapper.getWrappedRobot()));
    }

    /**
     * Returns hash code of this instance.
     *
     * @return hash code
     */
    public int hashCode() {
	return robot.hashCode();
    }

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (getName() + "; " + getOrientedPosition() + "; " + getEnergy());
    }

}
