/*
 * ShootingPath.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.List;

/**
 * Represents a shooting path as a list of actions leading to a shooting position for given robot.
 *
 * Note: Natural ordering of this class is incosistent with equals!
 *
 * @author yanas
 */
final class ShootingPath extends Path {

    /** target robot we want to shoot at */
    private final RobotWrapper targetRobot;
    /** number of points that might be gained by shooting */
    private final int pointsGainedByShooting;
    /** precalculated profitability factor */
    private final float profitabilityFactor;

    /**
     * Constructs a non-existing ShootingPath object.
     */
    public ShootingPath() {
	// call parent constructor
	super();

	// initialize remaining attributes
	targetRobot = null;
	pointsGainedByShooting = 0;
	profitabilityFactor = 0.0f;
    }

    /**
     * Constructs a ShootingPath object using the given parameters.
     * 
     * @param goal position of the goal in the world
     * @param actions a list of actions leading to the goal specified
     * @param targetRobot a target robot we want to shoot at
     * @param restRobotEnergy robot energy gained by resting
     */
    public ShootingPath(Position goal, List<BotAction> actions, RobotWrapper targetRobot,
	    int restRobotEnergy) {
	// call parent constructor
	super(goal, actions);

	// get current robot energy
	int robotEnergy = targetRobot.getEnergy();

	// follow the path step by step
	for (int i = 0; i < length(); i++) {
	    // decrement energy
	    robotEnergy--;

	    // if the robot didn't have enough energy
	    if (robotEnergy == 0) {
		// he must have rested so set the energy to rest energy
		robotEnergy = restRobotEnergy;
	    }
	}

	// calculate points gained by first shooting and robot update energy
	int pointsGainedByFirstShooting = (robotEnergy - 1) / 2;
	robotEnergy = robotEnergy - 1 - pointsGainedByFirstShooting;

	// if the robot didn't have enough energy
	if (robotEnergy == 0) {
	    // he must have rested so set the energy to rest energy
	    robotEnergy = restRobotEnergy;
	}

	// initialize remaining attributes
	this.targetRobot = targetRobot;
	this.pointsGainedByShooting = pointsGainedByFirstShooting + (robotEnergy - 1) / 2;
	this.profitabilityFactor = (float) pointsGainedByShooting
		/ ((length() + 1) * (length() + 1));
    }

    /**
     * Returns the target robot.
     *
     * @return the target robot
     */
    public RobotWrapper getTargetRobot() {
	return targetRobot;
    }

    /**
     * Returns the number of points gained by shooting.
     *
     * @return the number of points gained by shooting
     */
    public int getPointsGainedByShooting() {
	return pointsGainedByShooting;
    }

    /**
     * Returns the profitability factor based on points gained by shooting and path length.
     *
     * @return the profitability factor of this treasure path
     */
    public float getProfitabilityFactor() {
	return profitabilityFactor;
    }

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (super.toString() + "; " + getTargetRobot() + "; "
		+ getPointsGainedByShooting() + "; " + getProfitabilityFactor());
    }

}
