/*
 * ShootingPathFinder.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Robot.Direction;
import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Queue;

/**
 * Represents a shooting path finder based on A* algorithm.
 *
 * This implementations relies on an instance of the WorldWrapper class which should be
 * provided during construction of the ShootingPathFinder. The instance of the WorldWrapper
 * class represents a world in which the path finding is carried out. It also contains the
 * robot for which the path finding should be done.
 *
 * ShootingPathFinder does care about resting. It finds the shortest path (a sequence of
 * actions) between the robot's current position and the target position. The returned path
 * will probably be a bit longer than the one found by the PathFinder.
 *
 * @author yanas
 *
 * @see TreasurePathFinder
 * @see BatteryPathFinder
 * @see PathFinder
 *
 * @see ShootingPath
 */
final class ShootingPathFinder {

    /**
     * Represents a search node for path finding.
     *
     * These nodes are used to build a search tree. Each node has its predecessor and therefore
     * you can climb the tree back up to the root node. This is useful when building a list of
     * actions from the start node to the target node. Other attributes are used for heuristics.
     *
     * Note: Natural ordering of this class is incosistent with equals!
     *
     * @author  yanas
     */
    private static final class SearchNode implements Comparable<SearchNode> {

	/** this search node's predecessor */
	private final SearchNode predecessor;
	/** current oriented position */
	private final OrientedPosition orientedPosition;
	/** actions required to get from predecessor's oriented position to the current one */
	private final BotAction[] actions;
	/** total number of actions from start oriented position to the current one */
	private final int actionsCost;
	/** robot's remaining energy */
	private final int robotEnergy;

	/**
	 * Constructs a starting SearchNode object. This constructor should be called only once
	 * at the beginning of the path finding. Then you should use the second one as this
	 * constructor actually creates a root node.
	 *
	 * @param orientedPosition starting oriented position
	 * @param robotEnergy robot's remaining energy
	 */
	public SearchNode(OrientedPosition orientedPosition, int robotEnergy) {
	    // initialize all the attributes
	    this.predecessor = null;
	    this.orientedPosition = orientedPosition;
	    this.actions = null;
	    this.actionsCost = 0;
	    this.robotEnergy = robotEnergy;
	}

	/**
	 * Constructs a child SearchNode object. This constructor should be called when a root
	 * node has already been created. Use this constructor when expanding a node.
	 *
	 * @param predecessor direct predecessor of this node
	 * @param orientedPosition current oriented position
	 * @param actions actions required to get from predecessor's oriented position
	 * to the current one
	 * @param robotEnergy robot's remaining energy
	 */
	public SearchNode(SearchNode predecessor, OrientedPosition orientedPosition,
		BotAction[] actions, int robotEnergy) {
	    // initialize all the attributes
	    this.predecessor = predecessor;
	    this.orientedPosition = orientedPosition;
	    this.actions = actions;
	    this.actionsCost = predecessor.getActionsCost() + actions.length;
	    this.robotEnergy = robotEnergy;
	}

	/**
	 * Returns the predecessor of this node.
	 *
	 * @return predecessor of this node
	 */
	public SearchNode getPredecessor() {
	    return predecessor;
	}

	/**
	 * Returns the oriented position.
	 *
	 * @return the oriented position
	 */
	public OrientedPosition getOrientedPosition() {
	    return orientedPosition;
	}

	/**
	 * Returns the position.
	 *
	 * @return the position
	 */
	public Position getPosition() {
	    return orientedPosition.getPosition();
	}

	/**
	 * Returns the direction.
	 *
	 * @return the direction
	 */
	public Direction getDirection() {
	    return orientedPosition.getDirection();
	}

	/**
	 * Returns the array of actions.
	 *
	 * @return the array of actions
	 */
	public BotAction[] getActions() {
	    return actions;
	}

	/**
	 * Returns the actions cost needed to "get" from the root node to this one
	 *
	 * @return actions cost accumulated so far
	 */
	public int getActionsCost() {
	    return actionsCost;
	}

	/**
	 * Returns the robot's remaining energy.
	 *
	 * @return robot's remaining energy
	 */
	public int getRobotEnergy() {
	    return robotEnergy;
	}

	/**
	 * Compares this instance with another search node for order.
	 *
	 * @param searchNode another search node to be compared with this one
	 *
	 * @return an integer representing the relation between the nodes
	 */
	public int compareTo(SearchNode searchNode) {
	    return (getActionsCost() - searchNode.getActionsCost());
	}

	/**
	 * Returns string representation of this instance.
	 *
	 * @return important data converted to a string
	 */
	public String toString() {
	    return (getOrientedPosition() + "; " + getActionsCost() + "; " + getRobotEnergy());
	}

    }

    /** a world in which the search is carried out */
    private final WorldWrapper world;

    /** a priority queue of unexplored search nodes ordered by estimated cost */
    private final Queue<SearchNode> searchNodes = new PriorityQueue<SearchNode>();
    /** array of lowest actions costs per oriented position */
    private final int actionsCosts[][][];

    /**
     * Constructs an TreasurePathFinder object. An instance of the WorldWrapper class passed
     * to this constructor is used every time you want to find a path to a goal position and
     * cannot be modified.
     * 
     * @param world a world in which the search is carried out
     */
    public ShootingPathFinder(WorldWrapper world) {
	// save the world
	this.world = world;

	// create and initialize the actions cost array based on the world's dimensions
	actionsCosts = new int[world.getDimensionX()][world.getDimensionY()]
		[BotAction.values().length];
	clearActionsCosts();
    }

    /**
     * Finds the best shooting path based on the current world state.
     *
     * @param movingAllowed true if the robot is allowed to move around, false otherwise
     *
     * @return a ShootingPath object representing best shooting path
     */
    public ShootingPath findBestPath(boolean movingAllowed) {
	// prepare a list for shooting paths
	List<ShootingPath> shootingPaths = new ArrayList<ShootingPath>();

	// for each opponent robot in the world
	for (RobotWrapper opponentRobot : world.getOpponentRobots()) {
	    // find a shooting path to the given opponent
	    ShootingPath shootingPath = findPath(world.getNullBot(), opponentRobot.getPosition());

	    // if the path exists and moving around is allowed or the path does not require moving
	    if (shootingPath.exists() && (movingAllowed
		    || !shootingPath.getActions().contains(BotAction.STEP))) {
		// add the path to the list
		shootingPaths.add(shootingPath);
	    }
	}

	// return a non-existing path if no existing path has been found
	if (shootingPaths.isEmpty()) {
	    return new ShootingPath();
	}

	// finally sort the list with shooting paths
	Collections.sort(shootingPaths);

	return shootingPaths.get(0);
    }

    /**
     * Determines if anyone may shoot at us before the end of the game.
     *
     * @return true if anyone may shoot at us
     */
    public boolean mayAnyoneShootAtUs() {
	// for each opponent robot in the world
	for (RobotWrapper opponentRobot : world.getOpponentRobots()) {
	    // return true if a shooting path leading to our robot exists
	    if (findPath(opponentRobot, world.getNullBot().getPosition()).exists()) {
		return true;
	    }
	}

	return false;
    }

    /**
     * Returns string representation of this instance. For debugging purposes.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (world + "; " + searchNodes);
    }

    /**
     * Finds the shortest path between the associated robot's oriented position and the
     * best shooting position for the given target position. If there is not enough game
     * rounds to reach such position, the returned path won't exist.
     *
     * @param robot a robot for which the finding is done
     * @param target position to be shot at
     *
     * @return an instance of the ShootingPath class representing the result path
     */
    private ShootingPath findPath(RobotWrapper robot, Position target) {
	// return non-existing path if the goal is not accessible
	if (world.containsWall(target)) {
	    return new ShootingPath();
	}

	// create a starting search node
	SearchNode startNode = new SearchNode(robot.getOrientedPosition(), robot.getEnergy());

	// add the search node to the search nodes queue, set actions cost to zero
	searchNodes.add(startNode);
	setActionsCost(robot.getOrientedPosition(), 0);

	// while there are any search nodes in the search nodes queue
	while (!searchNodes.isEmpty()) {
	    // remove the best node from the nodes queue
	    SearchNode searchNode = searchNodes.remove();

	    // if the actions cost is higher or equal than the remaining rounds count
	    if (searchNode.getActionsCost() >= world.remainingRoundsCount()) {
		// break the loop to finish searching
		break;
	    }

	    // if the target can be shot
	    if (canShootAt(searchNode.getOrientedPosition(), target)) {
		// create an actions linked list
		LinkedList<BotAction> actions = new LinkedList<BotAction>();

		// while the node is not null and the node's actions are not null
		while (searchNode != null && searchNode.getActions() != null) {
		    // loop through all the actions in the array
		    for (int i = 0; i < searchNode.getActions().length; i++) {
			// add the action at the beginning of the actions list
			actions.add(i, searchNode.getActions()[i]);
		    }

		    // save this node's predecessor
		    searchNode = searchNode.getPredecessor();
		}

		// clear both nodes queue and actions costs
		searchNodes.clear();
		clearActionsCosts();

		return new ShootingPath(searchNode.getPosition(), actions,
			world.getRobotAt(target), world.getRestRobotEnergy());
	    }

	    // expand the current search node by stepping forward
	    expandSearchNode(searchNode, searchNode.getOrientedPosition().moveForward(),
		    BotAction.STEP, target);

	    // expand the current search node by turning left
	    expandSearchNode(searchNode, searchNode.getOrientedPosition().turnLeft(),
		    BotAction.LEFT, target);

	    // expand the current search node by turning right
	    expandSearchNode(searchNode, searchNode.getOrientedPosition().turnRight(),
		    BotAction.RIGHT, target);
	}

	// clear both nodes queue and actions costs
	searchNodes.clear();
	clearActionsCosts();

	return new ShootingPath();
    }

    /**
     * Returns true if the we can shoot at the robot at the target position from our current
     * oriented position, false otherwise.
     * 
     * @param orientedPosition current oriented position of the robot
     * @param targetPosition a position we want to shoot at
     * @return true/false specifying whether or not we can shoot
     */
    private boolean canShootAt(OrientedPosition orientedPosition, Position targetPosition) {
	// return true if the target robot stands in the same position as we do
	if (orientedPosition.getPosition().equals(targetPosition)) {
	    return true;
	}

	// prepare x and y offsets
	int xOffset = 0;
	int yOffset = 0;

	// which direction are we looking now?
	switch (orientedPosition.getDirection()) {
	    case UP:

		// return false if the x-coordinates don't match or the target stands behind
		if (orientedPosition.getPosition().getX() != targetPosition.getX()
			|| orientedPosition.getPosition().getY() < targetPosition.getY()) {
		    return false;
		}

		// set the y offset
		yOffset = -1;
		break;

	    case RIGHT:

		// return false if the y-coordinates don't match or the target stands behind
		if (orientedPosition.getPosition().getY() != targetPosition.getY()
			|| orientedPosition.getPosition().getX() > targetPosition.getX()) {
		    return false;
		}

		// set the x offset
		xOffset = 1;
		break;

	    case DOWN:

		// return false if the x-coordinates don't match or the target stands behind
		if (orientedPosition.getPosition().getX() != targetPosition.getX()
			|| orientedPosition.getPosition().getY() > targetPosition.getY()) {
		    return false;
		}

		// set the y offset
		yOffset = 1;
		break;

	    case LEFT:

		// return false if the y-coordinates don't match or the target stands behind
		if (orientedPosition.getPosition().getY() != targetPosition.getY()
			|| orientedPosition.getPosition().getX() < targetPosition.getX()) {
		    return false;
		}

		// set the x offset
		xOffset = -1;
		break;

	    // this should never happen
	    default:
		return false;
	}

	// offset the position until the wall is reached
	for (Position position = orientedPosition.getPosition().offset(xOffset, yOffset);
		!world.containsWall(position); position = position.offset(xOffset, yOffset)) {
	    // return boolean if there is a robot at the current position
	    if (world.containsRobot(position)) {
		return targetPosition.equals(position);
	    }
	}

	return false;
    }

    /**
     * Expands the given search node using given oriented position, actions and target.
     *
     * @param searchNode a search node to be expanded
     * @param orientedPosition oriented position we want to move to from the current position
     * @param action action required by the above movement
     * @param target a position we want to reach
     */
    private void expandSearchNode(SearchNode searchNode, OrientedPosition orientedPosition,
	    BotAction action, Position target) {
	// save the position we will move to by expanding the given seach node
	Position position = orientedPosition.getPosition();

	// if there is no wall and no robot or the position has not changed
	if ((!world.containsWall(position) && !world.containsRobot(position))
		|| position.equals(searchNode.getPosition())) {
	    // prepare an array of actions and robot energy
	    BotAction[] actions = {action};
	    int robotEnergy = searchNode.getRobotEnergy() - 1;

	    // if the robot didn't have enough energy to carry out the action
	    if (robotEnergy < 1) {
		// update the array of actions
		actions = new BotAction[2];
		actions[0] = BotAction.REST;
		actions[1] = action;

		// set the robot energy to rest energy
		robotEnergy = world.getRestRobotEnergy();
	    }

	    // calculate the total actions cost from start to this oriented position
	    int totalActionsCost = searchNode.getActionsCost() + actions.length;

	    // if the actions cost is lower than the best one stored so far
	    if (totalActionsCost < getActionsCost(orientedPosition)) {
		// create and add a new node to the search nodes queue
		searchNodes.add(new SearchNode(searchNode, orientedPosition, actions,
			robotEnergy));

		// set new actions cost for this oriented position calculated earlier
		setActionsCost(orientedPosition, totalActionsCost);
	    }
	}
    }

    /**
     * Returns the action cost for the given oriented position.
     *
     * @param orientedPosition an oriented position we want to know about
     *
     * @return a non-zero integer when the oriented position has already been visited,
     * short integer max otherwise
     */
    private int getActionsCost(OrientedPosition orientedPosition) {
	// get position and direction
	Position position = orientedPosition.getPosition();
	Direction direction = orientedPosition.getDirection();

	return actionsCosts[position.getX()][position.getY()][direction.ordinal()];
    }

    /**
     * Sets action cost for the given oriented position.
     *
     * @param orientedPosition an oriented position for which the actions is to be set
     * @param actionsCost new actions cost to be set
     */
    private void setActionsCost(OrientedPosition orientedPosition, int actionsCost) {
	// get position and direction
	Position position = orientedPosition.getPosition();
	Direction direction = orientedPosition.getDirection();

	// set the actions cost
	actionsCosts[position.getX()][position.getY()][direction.ordinal()] = actionsCost;
    }

    /**
     * Clears the actions costs array by filling each cell with short integer max value.
     * This value means that an oriented position has not yet been visited.
     */
    private void clearActionsCosts() {
	// loop through all the arrays in the actions costs array
	for (int i = 0; i < world.getDimensionX(); i++) {
	    // loop through all the sub-arrays in the actions costs array
	    for (int j = 0; j < world.getDimensionY(); j++) {
		// fill each array with max int value as the highest actions cost
		Arrays.fill(actionsCosts[i][j], Short.MAX_VALUE);
	    }
	}
    }

}
