/*
 * TreasurePath.java
 *
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.Strategy.BotAction;

import java.util.List;

/**
 * Represents a treasure path as a list of actions leading to a given treasure.
 *
 * Note: Natural ordering of this class is incosistent with equals!
 *
 * @author yanas
 */
final class TreasurePath extends Path {

    /** value of the treasure */
    private final int treasureValue;
    /** cumulative value of the treasure around this one */
    private final float cumulativeTreasureValue;
    /** precalculated profitability factor */
    private final float profitabilityFactor;

    /**
     * Constructs a non-existing TreasurePath object.
     */
    public TreasurePath() {
	// call parent constructor
	super();

	// initialize remaining attributes
	treasureValue = 0;
	cumulativeTreasureValue = 0.f;
	profitabilityFactor = 0.0f;
    }

    /**
     * Constructs a TreasurePath object using the given parameters.
     *
     * @param goal position of the goal in the world
     * @param actions a list of actions leading to the goal specified
     * @param treasureValue actual value of the treasure
     * @param cumulativeTreasureValue cumulative value of treasures around this one
     */
    public TreasurePath(Position goal, List<BotAction> actions, int treasureValue,
	    float cumulativeTreasureValue) {
	// call parent constructor
	super(goal, actions);

	// initialize remaining attributes
	this.treasureValue = treasureValue;
	this.cumulativeTreasureValue = cumulativeTreasureValue;
	this.profitabilityFactor = cumulativeTreasureValue / (length() + 1);
    }

    /**
     * Returns the treasure value.
     *
     * @return the treasure value
     */
    public int getTreasureValue() {
	return treasureValue;
    }

    /**
     * Returns the cumulative treasure value.
     *
     * @return the cumulative treasure value
     */
    public float getCumulativeTreasureValue() {
	return cumulativeTreasureValue;
    }

    /**
     * Returns the profitability factor based on cumulative treasure value and path length.
     *
     * @return the profitability factor of this treasure path
     */
    public float getProfitabilityFactor() {
	return profitabilityFactor;
    }

    /**
     * Returns string representation of this instance.
     *
     * @return important data converted to a string
     */
    public String toString() {
	return (super.toString() + "; " + getTreasureValue() + "; "
		+ getCumulativeTreasureValue() + "; " + getProfitabilityFactor());
    }

}
