/*
 * WorldWrapper.java
 *
 * @author network_gauner
 * @author yanas
 * @version 21st May 2006, 17:43
 */

package cz.muni.fi.fibot.nullbot;

import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Robot;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * A wrapper around the World class providing more functionality.
 * 
 * @author network_gauner
 * @author yanas
 */
final class WorldWrapper {

    /** wrapped world */
    private final World world;
    /** our robot */
    private final RobotWrapper nullbot;

    /** a set of all robots */
    private final Set<RobotWrapper> robots = new HashSet<RobotWrapper>();
    /** a set of accessible opponent robots */
    private final Set<RobotWrapper> opponentRobots = new HashSet<RobotWrapper>();

    /** a set of accessible treasures */
    private final Set<Position> treasures = new HashSet<Position>();
    /** a list of accessible batteries */
    private final Set<Position> batteries = new HashSet<Position>();

    /** a set of positions that might be blocked by opponent robots in this and next round */
    private final Set<Position> opponentRobotObstacles = new HashSet<Position>();

    /**
     * Constructs a WorldWrapper object around the given world and our robot.
     *
     * @param world a world to be wrapped
     * @param nullbot our robot
     */
    public WorldWrapper(World world, RobotWrapper nullbot) {
	// save world and our robot
        this.world = world;
	this.nullbot = nullbot;

	// prepare a general path finder and a set of opponent positions
	PathFinder pathFinder = new PathFinder(this);
	Set<Position> opponentPositions = new HashSet<Position>();

	// for each robot in the world
	for (Robot robot : world.getRobots()) {
	    // if the the current robot is equal to our robot
            if (robot.equals(nullbot.getWrappedRobot())) {
		// add our robot to the set of robots
                robots.add(nullbot);
	    }
	    else {
		// create a robot wrapper for the current robot
		RobotWrapper robotWrapper = new RobotWrapper(robot);

		// add it as a robot as well as a opponent robot
                robots.add(robotWrapper);
		opponentRobots.add(robotWrapper);

		// add the robot position to the corresponding set
		opponentPositions.add(robotWrapper.getPosition());
            }
        }

	// loop through all the x-coordinates
	for (int x = 0; x < world.getDimensionX(); x++) {
	    // loop through all the y-coordinates
            for (int y = 0; y < world.getDimensionY(); y++) {
		// prepare position for current coordinates
                Position position = new Position(x, y);

		// if the position is accessible
		if (pathFinder.pathExists(nullbot.getOrientedPosition(), position)) {
		    // if there is a treasure
		    if (containsTreasure(position)) {
			// add the treasure to the treasures set
			treasures.add(position);
		    }
		    // else if there is a battery
		    else if (containsBattery(position)) {
			// add the battery to the batteries set
			batteries.add(position);
		    }
		}
            }
        }

	// if no opponent is accessible
	if (!pathFinder.anyPathExists(nullbot.getOrientedPosition(), opponentPositions)) {
	    // remove all opponent robots
	    opponentRobots.clear();
	}
    }

    /**
     * Returns the world associated with this wrapper.
     *
     * @return the world associated with this wrapper
     */
    public World getWrappedWorld() {
        return world;
    }

    /**
     * Returns our robot.
     *
     * @return our robot
     */
    public RobotWrapper getNullBot() {
	return nullbot;
    }

    /**
     * Returns the x dimension of the world.
     * 
     * @return the x dimension of the world
     */
    public int getDimensionX() {
        return world.getDimensionX();
    }

    /**
     * Returns the y dimension of the world.
     * 
     * @return the y dimension of the world
     */
    public int getDimensionY() {
        return world.getDimensionY();
    }

    /**
     * Returns the maximum energy a robot can have.
     * 
     * @return the maximum energy
     */
    public int getMaxRobotEnergy() {
        return world.getMaxRobotEnergy();
    }

    /**
     * Returns the amount of energy gained after rest action.
     * 
     * @return the amount of energy gained after rest action
     */
    public int getRestRobotEnergy() {
        return world.getRestRobotEnergy();
    }

    /**
     * Returns the total number of game rounds.
     *
     * @return the total number of game rounds
     */
    public int getRoundsCount() {
        return world.getRoundsCount();
    }

    /**
     * Returns the ordinal of current game round.
     *
     * @return the ordinal of current game round
     */
    public int getCurrentRound() {
        return (world.getRound() + 1);
    }

    /**
     * Returns the number of remaining game rounds.
     *
     * @return the number of remaining game rounds
     */
    public int remainingRoundsCount() {
        return (world.getRoundsCount() - world.getRound());
    }

    /**
     * Returns true if the current round is the last round, false otherwise.
     *
     * @return true if the current round is the last round
     */
    public boolean isLastRound() {
	return (getCurrentRound() == getRoundsCount());
    }

    /**
     * Returns true if the given position is in the world, false otherwise.
     *
     * @return true if the given position is in the world
     */
    public boolean isInWorld(Position position) {
        return (position.getX() >= 0 && position.getX() < world.getDimensionX()
		&& position.getY() >= 0 && position.getY() < world.getDimensionY());
    }
    
    /**
     * Returns true if there is a wall at the given position, false otherwise.
     *
     * @return true if there is a wall at the given position
     */
    public boolean containsWall(Position position) {
        return (!isInWorld(position) || world.containsWall(position.getX(), position.getY())
                || containsInactiveRobot(position));
    }

    /**
     * Returns true if there is an obstacle at the given position, false otherwise.
     *
     * @return true if there is an obstacle at the given position
     */
    public boolean containsObstacle(Position position) {
	return (containsWall(position)|| opponentRobotObstacles.contains(position));
    }

    /**
     * Returns true if there is a robot at the given position, false otherwise.
     *
     * @return true if there is a robot at the given position
     */
    public boolean containsRobot(Position position) {
        return (getRobotAt(position) != null);
    }

    /**
     * Returns true if there is an inactive robot at the given position, false otherwise.
     *
     * @return true if there is an inactive robot at the given position
     */
    public boolean containsInactiveRobot(Position position) {
        return (containsRobot(position) && getRobotAt(position).getEnergy() == 0);
    }

    /**
     * Returns true if there is a treasure at the given position, false otherwise.
     *
     * @return true if there is a treasure at the given position
     */
    public boolean containsTreasure(Position position) {
        return (isInWorld(position) && world.containsTreasure(position.getX(), position.getY()));
    }

    /**
     * Returns true if there is a battery at the given position, false otherwise.
     *
     * @return true if there is a battery at the given position
     */
    public boolean containsBattery(Position position) {
        return (isInWorld(position) && world.containsBattery(position.getX(), position.getY()));
    }

    /**
     * Returns the robot standing at the given position or null if there is no such robot.
     *
     * @return an instance of RobotWrapper class or null
     */
    public RobotWrapper getRobotAt(Position position) {
	// for each robot
	for (RobotWrapper robot : robots) {
	    // return the robot if the positions equal
	    if (position.equals(robot.getPosition())) {
		return robot;
	    }
	}

	return null;
    }

    /**
     * Returns a set of robots.
     * 
     * @return a set of robots
     */
    public Set<RobotWrapper> getRobots() {
	return Collections.unmodifiableSet(robots);
    }

    /**
     * Returns a set of accessible opponent robots.
     * 
     * @return a set of accessible opponent robots
     */
    public Set<RobotWrapper> getOpponentRobots() {
	return Collections.unmodifiableSet(opponentRobots);
    }

    /**
     * Returns value of a treasure at the given position or 0 if there is no such treasure.
     * 
     * @return treasure value or 0 if the treasure is not present
     */
    public int getTreasureValue(Position position) {
        return (containsTreasure(position) ? world.getTreasureValue(position.getX(),
		position.getY()) : 0);
    }

    /**
     * Returns a cumulative treasure value around the given position.
     *
     * @param position centre position of the area to be explored
     *
     * @return cumulative treasure value in the area
     */
    public float cumulativeTreasureValue(Position position) {
	// initialize cumulative treasure value to zero
	float cumulativeTreasureValue = 0.0f;

	// loop through all the x values
	for (int x = position.getX() - 1; x <= position.getX() + 1; x++) {
	    // loop through all the y values
	    for (int y = position.getY() - 1; y <= position.getY() + 1; y++) {
		// get current treasure position, add treasure value divided by a quotient
		Position treasurePosition = new Position(x, y);
		cumulativeTreasureValue += (float) getTreasureValue(treasurePosition)
			/ (Position.manhattanDistance(position, treasurePosition) + 1);
	    }
	}

	return cumulativeTreasureValue;
    }

    /**
     * Returns a set of available and accessible treasures currently in the world.
     *
     * @return an unmodifiable set of treasures
     */
    public Set<Position> getTreasures() {
	// iterate over the set of treasures
        for (Iterator<Position> i = treasures.iterator(); i.hasNext(); ) {
	    // save the current position
            Position position = i.next();

	    // if there is no treasure or the position is blocked by an inactive robot
	    if (!containsTreasure(position) || containsInactiveRobot(position)) {
		// remove the treasure from the set
                i.remove();
            }
        }

	return Collections.unmodifiableSet(treasures);
    }

    /**
     * Returns a set of available and accessible batteries currently in the world.
     *
     * @return an unmodifiable set of batteries
     */
    public Set<Position> getBatteries() {
	// iterate over the list of batteries
        for (Iterator<Position> i = batteries.iterator(); i.hasNext(); ) {
	    // if the battery is blocked by an inactive robot
            if (containsInactiveRobot(i.next())) {
		// remove if from the list
                i.remove();
            }
        }

	return Collections.unmodifiableSet(batteries);
    }

    /**
     * Returns true if a collision may occur in the next round, false otherwise.
     *
     * @return true if a collision may occur
     */
    public boolean mayCollisionOccur() {
	// determine opponent robot obstacles first
	determineOpponentRobotObstacles();

	// determine if the collision may occur or not
	boolean mayCollisionOccur = opponentRobotObstacles.contains(
		nullbot.getOrientedPosition().moveForward().getPosition());

	// clear opponent robot obstacles
	clearOpponentRobotObstacles();

	return mayCollisionOccur;
    }

    /**
     * Determines opponent robot obstacles based on the current opponent robot positions.
     */
    public void determineOpponentRobotObstacles() {
	// for each opponent robot
	for (RobotWrapper opponentRobot : getOpponentRobots()) {
	    // mark the opponent robot's position and get the position in front of it
	    opponentRobotObstacles.add(opponentRobot.getPosition());
	    Position position = opponentRobot.getOrientedPosition().moveForward().getPosition();

	    // if the position is not occupied and the robot has enough energy to make a step
	    if (!position.equals(nullbot.getPosition()) && opponentRobot.getEnergy() > 1) {
		// mark the position as a robot obstacle
		opponentRobotObstacles.add(position);
	    }
	}
    }

    /**
     * Clears opponent robot obstacles determined earlier.
     */
    public void clearOpponentRobotObstacles() {
	opponentRobotObstacles.clear();
    }

}
