package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

/**
 * Třída implemetující vlastní strategii
 */
public class RossumsUniversalRobot implements Strategy {

    private World world;        // svět
    private Robot robot;        // můj robot
    
    private Battery battery;    // baterie
    
    private Tarjan tarjanRobot;                         // tarjan pro mého robota
    private TarjanWithEnergy tarjanWithEnergyRobot;     // tarjan pro mého robota pocitajici s energii
    
    private Map<Robot, Tarjan> tarjanEnemies = new HashMap<Robot, Tarjan>();        // tarjan pro ostatní roboty
    private Map<Robot, TarjanWithEnergy> tarjanWithEnergyEnemies = new HashMap<Robot, TarjanWithEnergy>();    // tarjan pro ostatní roboty pocitajici s energii
    
    private Treasure treasure;              //poklady
    private Sniper sniper;                  //střelba
    private Rest rest;                      //odpočívání před koncem
    
    private BotAction nextAction;           //akce kterou provedu v dalším kole
    
    //promene pro detekci uvaznuti
    private int collisionCounter = 0;
    private BotAction lastAction;
    private Position lastPosition;
   
    /**
     * Konstruktor strategie
     */
    public RossumsUniversalRobot() {        
       
    }
    
    /**
     * Inicializace na začátku každé hry
     */
    public void initialize() {
        
        battery = new Battery(world);
       
        tarjanRobot = new Tarjan(world);                  // tarjan pro mého robota
        tarjanWithEnergyRobot = new TarjanWithEnergy(world, robot, tarjanRobot, battery);   // tarjan pro mého robota e energii
        
        for (Robot enemy : world.getRobots()) {
            if (enemy != robot)  {                             // já nejsem nepřátel
                tarjanEnemies.put(enemy, new Tarjan(world));   // tarjani mých nepřátel  
                tarjanWithEnergyEnemies.put(enemy, new TarjanWithEnergy(world, enemy, tarjanEnemies.get(enemy),  battery));   // tarjani mých nepřátel  
        
            }
        }             
            
        treasure = new Treasure(world, robot, tarjanWithEnergyRobot, tarjanWithEnergyEnemies,  battery);      //inicializace pokladů
        sniper = new Sniper(world, robot, tarjanWithEnergyRobot, tarjanWithEnergyEnemies);   //inicializace střelby
        rest = new Rest(world, robot, tarjanWithEnergyRobot, tarjanWithEnergyEnemies, battery);   //inicializace odpočívání před koncem
    }
    
    /**
     * Nastaví svět
     */
    public void setWorld(World w) {
        this.world = w;
    }
    
    /**
     * Nastaví mého robota
     */
    public void setRobot(Robot r) {
        this.robot = r;   
    }
    
    /**
     * Rozhodovací funkce naší strategie
     *
     * @return
     *      akce, kterou má můj robot provést
     */
    public Strategy.BotAction decide() {
        try {
            // BEGIN - inicializace v káždém kole  
            tarjanRobot.compute(new Position(robot)); // prepočítej tarjana pro mého robota   

            for (Robot enemy : tarjanEnemies.keySet()) { // prepočítej tarjany pro nepřátele            
                tarjanEnemies.get(enemy).compute(new Position(enemy));  
            }  
            //END - inicializace v káždém kole


            // BEGIN - predvidatelna kolize
            Position next = new Position(robot.getPositionX(),robot.getPositionY());
            if (robot.getDirection() == Robot.Direction.DOWN) next.setY(next.getY()+1); 
            else if (robot.getDirection() == Robot.Direction.UP) next.setY(next.getY()-1);
            else if (robot.getDirection() == Robot.Direction.RIGHT) next.setX(next.getX()+1);
            else next.setX(next.getX()-1);
            if (next.isCorrect(world) && world.containsRobot(next.getX(), next.getY())) {
                Robot collisionRobot = world.getRobotOn(next.getX(), next.getY());
                if (sniper.stayInXY(collisionRobot) > 0) {
                   tarjanRobot.collision=true;
                   tarjanRobot.compute(new Position(robot));
                   //Log.print(Log.Level.LOW, "COLLISION 100% ON");
                }
           }
           // END - predvidatelna kolize 



            double treasureAttract = treasure.mostAttractValue(); //atraktivita brani pokladu
            double sniperAttract = sniper.mostAttractValue();  //atraktivitra strileni
            double restAttract = rest.mostAttractValue(); //atraktivita odpocivani pred koncem

            //Log.print(Log.Level.LOW, "ATTRACT:  T=" + treasureAttract + "  S=" +sniperAttract + "  R=" + ((restAttract==Integer.MAX_VALUE)?"MAX_INT":restAttract));

            if (treasureAttract >= sniperAttract && treasureAttract >= restAttract) {
                //nejvýhodnější je brát poklady            
                nextAction = treasure.mostAttractAction();
                //Log.print(Log.Level.LOW, "ATTRACT:  winner=TREASURE");
            } else if (sniperAttract >= restAttract)  {
                //nejvýhodnější je střílet           
                nextAction = sniper.mostAttractAction(); 
                //Log.print(Log.Level.LOW, "ATTRACT:  winner=SNIPER");                
            }  else { 
                //nejvýhodnější je odpočívat            
                nextAction = rest.mostAttractAction();
                //Log.print(Log.Level.LOW, "ATTRACT:  winner=REST"); 
            } 


            //detekce kolizi
            if (lastAction==Strategy.BotAction.STEP && lastPosition.getX()==robot.getPositionX() && lastPosition.getY()==robot.getPositionY()) {
                collisionCounter++;
                //Log.print(Log.Level.LOW, "COLLISION:  collisionCounter=" + collisionCounter);
                if (collisionCounter==4) {
                    tarjanRobot.collision=true;
                    //Log.print(Log.Level.LOW, "COLLISION PROBABLE ON");
                }
            } else if (lastAction!=Strategy.BotAction.REST)  {
                tarjanRobot.collision=false;
                collisionCounter=0;
            }

            lastAction=nextAction;
            lastPosition=new Position(robot.getPositionX(), robot.getPositionY());
            return nextAction; 
        } catch (Exception e) {
            return Strategy.BotAction.REST;
        }
    }

}