package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.Strategy;
import cz.muni.fi.fibot.World;

import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Map;
import java.util.HashMap;


/**
 *  Třída počitající nejkratší cesty z výchozí pozice do všech ostatních 
 */
public class Tarjan {
    
    public static final int NO_PATH = 999; // délka nedosažitelné cesty
    public boolean collision=false;
    
    private int dimX, dimY;         // rozměr hřiště
    private int roundsToEnd;        // počet kol do konce
    private World world;            // zdůvodu použití world.containsWall

    private int[] start;            // startovní pozice tarjana
    private int[][][] dirdist;      // nejkratší vzdálenosti
    private Integer[][][] prev;     // předchudce na nejkratší cestě
    
    /**
     * Konstruktor, pouze inicializuje svět
     *
     * @param world
     *      svět na kterém se bude počítat
     */  
    public Tarjan(World world) {
        
        // inicializace údaju o světě
        this.world = world;
        dimX = world.getDimensionX();
        dimY = world.getDimensionY();
         
        // inicializace polí
        dirdist = new int[dimX][dimY][4];
        prev = new Integer[dimX][dimY][4];  
    }
    
    
    /**
     * Převodní funkce, kóduje (x,y,d,value -> Integer) pro indexaci containeru waiting
     *
     * @param x
     *      x-ová souřadnice
     * @param y   
     *      y-ová souřadnice
     * @param d   
     *      číselná reprezentace směru 
     * @param value   
     *      délka nejkratší zatím nalezené cesty
     * @return 
     *      index do containeru waiting
     */  
    private Integer codeWaitingIndex(int x, int y, int d, int value) {
        return d | y<<2 | x<<8 | value<<14;
    }
    
    
    /**
     * Převodní funkce, dekóduje (Integer -> x,y,d) pro získání hodnot z containeru waiting
     *
     * @param index
     *      index do datové struktury waiting
     * @return 
     *      pozici ve tvaru int [] {x,y,d}
     */  
    private int[] decodeWaitingIndex(Integer index) {        
        return new int[] { (index>>8) % 64, (index>>2) % 64, index % 4 };   
    }
    
    
    /**
     * Přepočítá minimální cesty pro zadanou výchozí pozici
     *
     * @param startPosition
     *      výchozí pozice
     */  
    public void compute(Position startPosition) {
        
        int x, y, d;        //aktualní pozice po složkách
        int[] xyd;          //aktualní pozice v poli
        Integer actDist;    //aktualní vzdálenost
        
        //vynulování polí
        for (int ix=0; ix < dimX; ix++) {
            for (int iy=0; iy < dimY; iy++) {
                for (int id=0; id < 4; id++) {
                        dirdist[ix][iy][id] = NO_PATH;
                }
            }
        }
        
        start = new int[] {startPosition.getX(), startPosition.getY(), Position.dirToInt(startPosition.getD())};   
        roundsToEnd = world.getRoundsCount() - world.getRound();
                      
        SortedSet<Integer> waiting = new TreeSet<Integer>(); 
        
        dirdist[startPosition.getX()][startPosition.getY()][Position.dirToInt(startPosition.getD())] = 0;
        
        waiting.add(codeWaitingIndex(startPosition.getX(), startPosition.getY(), Position.dirToInt(startPosition.getD()), 0));
        
        while (!waiting.isEmpty()) {
           
            xyd = decodeWaitingIndex(waiting.first());
            x = xyd[0];  y = xyd[1];  d = xyd[2];
            waiting.remove(waiting.first());
            
            actDist = dirdist[x][y][d];
            
            if (actDist>roundsToEnd) break; // omezení hloubky stromu    
            
            testBetterWay(x, y, (d + 1) % 4, actDist, waiting, xyd);
            testBetterWay(x, y, (d + 3) % 4, actDist, waiting, xyd);
            
            if (collision && actDist==0) continue;
                
            if ((d == 0) && testIsFree(x, y-1)) {
                testBetterWay(x, y-1, d, actDist, waiting, xyd);
            }
            if ((d == 1) && testIsFree(x+1, y)) {
                testBetterWay(x+1, y, d, actDist, waiting, xyd);
            }
            if ((d == 2) && testIsFree(x, y+1)) {
                testBetterWay(x, y+1, d, actDist, waiting, xyd);
            }
            if ((d == 3) && testIsFree(x-1, y)) {
                testBetterWay(x-1, y, d, actDist, waiting, xyd);
            }
            
        }        
       
    }
    
      
    /**
     * Pomocná funkce pro výpočet minimálních cest.
     */  
    private void testBetterWay(int x1, int y1, int d1, int actDist, SortedSet<Integer> waiting, int[] xyd) {
        if (dirdist[x1][y1][d1] > actDist + 1) {
            if (dirdist[x1][y1][d1] == NO_PATH) {
                waiting.add(codeWaitingIndex(x1, y1, d1, dirdist[x1][y1][d1])); 
            }
            waiting.remove(codeWaitingIndex(x1, y1, d1, dirdist[x1][y1][d1]));
            dirdist[x1][y1][d1] = actDist + 1;
            waiting.add(codeWaitingIndex(x1, y1, d1, dirdist[x1][y1][d1]));
           
            prev[x1][y1][d1] = codeWaitingIndex(xyd[0], xyd[1], xyd[2], 0);
        }
    }
    
    
    /**
     * Zkontroluje, jestli je dané políčko na mapě volné
     *
     * @param x
     *      x-ová souřadnice políčka
     * @param y
     *      y-ová souřadnice políčka
     * @return 
     *      true/false je/není políčko volné
     */  
    private boolean testIsFree(int x, int y) {
        if (dimX <= x || dimY <= y || x < 0 || y < 0 || world.containsWall(x, y)) return false; //offside nebo zeď
        if (world.containsRobot(x, y) && world.getRobotOn(x, y).getEnergy()==0) return false; //mrtvola
        return true;
    }
    

    /**
     * Vrací délku nejkratší cesty z výchozí pozice do zadaná cílové pozice
     *
     * @param position
     *      cílová pozice
     * @return 
     *      délka nejkratší cesty
     */  
    public int distance(Position position) {    
        if (position.getD() == null) {
            int posX = position.getX();
            int posY = position.getY();            
            return Math.min(Math.min(dirdist[posX][posY][0], dirdist[posX][posY][1]), Math.min(dirdist[posX][posY][2],dirdist[posX][posY][3]));
        } else {
            return dirdist[position.getX()][position.getY()][Position.dirToInt(position.getD())];
        }               
    }    
    
        
    /**
     * Vrací první akci pro rekonstrukci minimální cesty z výchozí pozice do zadaná cílové pozice
     *
     * @param position
     *      cílová pozice
     * @return 
     *      první akce kterou musím udělat
     */  
    public Strategy.BotAction action(Position position) {
        
        int posX = position.getX();
        int posY = position.getY();
        Robot.Direction posD = position.getD();
        
        if (posD == null) {
                        
            if (dirdist[posX][posY][0] == distance(new Position(posX, posY))) 
                return action(new Position(posX, posY, Robot.Direction.UP));
            else if (dirdist[posX][posY][1] == distance(new Position(posX, posY))) 
                return action(new Position(posX, posY, Robot.Direction.RIGHT));
            else if (dirdist[posX][posY][2] == distance(new Position(posX, posY))) 
                return action(new Position(posX, posY, Robot.Direction.DOWN));
            else 
                return action(new Position(posX, posY, Robot.Direction.LEFT));
        
        } else {
                
            if (dirdist[posX][posY][Position.dirToInt(posD)] == 0) {
                return null;
            }

            int[] xyd = {posX,posY,Position.dirToInt(posD)};
            Integer target;

            while (dirdist[xyd[0]][xyd[1]][xyd[2]] != 1) {
                target = prev[xyd[0]][xyd[1]][xyd[2]];
                xyd = decodeWaitingIndex(target);            
            }

            if (xyd[2] == start[2]) {
                return Strategy.BotAction.STEP;
            } else if (xyd[2] == (start[2] + 1) % 4) {
                return Strategy.BotAction.RIGHT;
            } else {
                return Strategy.BotAction.LEFT;
            } 

        }
         
    }       
    

}



