package cz.muni.fi.fibot.rossumsuniversalrobot;

import cz.muni.fi.fibot.Robot;
import cz.muni.fi.fibot.World;
import cz.muni.fi.fibot.Strategy;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;


/**
 * Třída pro "substrategie" týkající se pokladů
 */
public class Treasure {
    
    World world;        // svět
    Robot robot;        // můj robot
    Battery battery;    // baterky
    
    TarjanWithEnergy tarjanWithEnergyRobot; // tarjan mého robota
    Map<Robot,TarjanWithEnergy> tarjanWithEnergyEnemies; //tarjani nepřátel
    
    private static Map<Integer, Tarjan> tarjanTreasures;  // zatim spocitani trajani pro poklady  
    
    private Position lastBestPosition; // nevýhodnějsí pozice, kterou našlo poslední volání mostAttractValue()
    private Position lastBatteryPosition; // baterie na nejvyhodnejsi ceste, kterou našlo poslední volání mostAttractValue()
    
    /**
     * Seznam (List) pozic pokladů
     */
    List<Position> treasures = new ArrayList<Position>();    
    
    /**
     * Konstruktor, inic. svět, robota a tarjany
     *
     * @param world
     *      svět
     * @param robot
     *      náš robot
     * @param tarjanRobot
     *      tarjan mého robota
     * @param tarjanEnemies
     *      tarjani nepřátel
     */
    public Treasure(World world, Robot robot, TarjanWithEnergy tarjanWithEnergyRobot, Map<Robot,TarjanWithEnergy> tarjanWithEnergyEnemies, Battery battery) {  
        this.world = world;
        this.robot = robot;
        this.tarjanWithEnergyRobot = tarjanWithEnergyRobot;
        this.tarjanWithEnergyEnemies = tarjanWithEnergyEnemies;
        this.battery=battery;
        
        tarjanTreasures = new HashMap<Integer, Tarjan>();
              
        for (int x = 0; x < world.getDimensionX(); x++) {
            for (int y = 0; y < world.getDimensionY(); y++) {
                if (world.containsTreasure(x, y)) treasures.add(new Position(x, y));
            }
        }
    }
    
    /**
     * Odstraní z treasures poklady, které z mapy zmizely nebo jsou pro našeho robota daleko (nedosažitelné)
     * 
     * @param tarjanRobot
     *      tarjan pro našeho robota
     */
    private void treasureControl() {
        Set<Position> delete = new HashSet<Position>();
        
        for (Position treasure : treasures) {
            if (!world.containsTreasure(treasure.getX(), treasure.getY()) || tarjanWithEnergyRobot.distance(treasure) > world.getRoundsCount() - world.getRound()) { 
                  delete.add(treasure);          
            }
        }
        for (Position treasure : delete) {
            treasures.remove(treasure);
        }        
    }
    
    /**
     * Vrací atraktivitu nejatraktivnějšího pokladu
     * 
     * @return Double
     *      koeficient atraktivnosti
     */
    public double mostAttractValue() {
       
        treasureControl(); //odstraň, co už není na mapě, nebo k tomu už nestihnu dojít
        
        Map<Position, Double> attract = new HashMap<Position, Double>(); 
        double actualAttract = 0;
        double maxAttract = 0;
        
        int robotDist, enemyDist, treasureValue;
        
        //spočítej atraktivnost jednotlivých pokladů
        for (Position treasure : treasures) {
            //Log.print(Log.Level.MEDIUM, "T::    position=" + treasure + " value=" + world.getTreasureValue(treasure.getX(),treasure.getY()));
            
            robotDist = tarjanWithEnergyRobot.distanceWithBattery(treasure);
            treasureValue = world.getTreasureValue(treasure.getX(), treasure.getY()); 
            
            actualAttract = (double) treasureValue / (robotDist + 1);
            
            for (Robot enemy : tarjanWithEnergyEnemies.keySet()) {
                enemyDist = tarjanWithEnergyEnemies.get(enemy).distance(treasure);              
                if (robotDist > enemyDist) {
                    actualAttract *= Math.sqrt((double) enemyDist / robotDist);                    
                }
            }
            
            //Log.print(Log.Level.MEDIUM, "T::        firstAttract: " + actualAttract);
            
            if (Parameters.TREASURE_DEEP > 1 && actualAttract*Parameters.TREASURE_DEEP > maxAttract) { //OPTIMALIACE aby se nesazil zlepsit to co uz nebude nejlepsi
                Map<Integer, Position> newTakenTreasures = new HashMap<Integer, Position>();
                newTakenTreasures.put(0,treasure);
                actualAttract +=  bestAttractFromPosition(newTakenTreasures, robotDist+1, 0);
            }
            
            //Log.print(Log.Level.MEDIUM, "T::        finalAttract: " + actualAttract);
                       
            if (maxAttract < actualAttract) {
                lastBestPosition = treasure;
                lastBatteryPosition = tarjanWithEnergyRobot.targetBattery;
                maxAttract = actualAttract;
            }
        }      
        
        //Log.print(Log.Level.MEDIUM, "T::    maxAttract: " + maxAttract +" position: " + lastBestPosition);
        if (lastBatteryPosition!=null) {
            //Log.print(Log.Level.MEDIUM, "T::    battery: " + lastBatteryPosition);
        }
        
        return maxAttract;
    }   
    
    
     /**
     * Vrací akci, kterou mám provést pokud chci využít nevýhodnějsí pozice, kerou našlo poslední volání mostAttractValue()
     * 
     * @return Strategy.BotAction
     *      akce která se má provést
     */
    public Strategy.BotAction mostAttractAction() {
        if (lastBatteryPosition==null) {
            if (tarjanWithEnergyRobot.action(lastBestPosition) != null) { // pokud nejsem na místě
                return tarjanWithEnergyRobot.action(lastBestPosition);
            } else { 
                return Strategy.BotAction.TAKE;
            }
        } else {
            if (tarjanWithEnergyRobot.action(lastBatteryPosition) != null) {
                if (robot.getEnergy()<2 && world.containsBattery(robot.getPositionX(),robot.getPositionY()))   //kdyz umiram sem na baterce tak at dam RECHARGE misto planovaneho REST                  
                    return Strategy.BotAction.RECHARGE;               
                else          
                    return tarjanWithEnergyRobot.action(lastBatteryPosition);
            } else { 
                return Strategy.BotAction.RECHARGE;
            }
        } 
    }     
    
    /**
     * Počítá atraktivnost pokladu ze zadané pozice vzhledem k aktuální pozici
     *
     * @param position 
     *      pozice, ze které se má atraktivnost počítat
     * @return Double
     *      koeficient atraktivnosti
     */
    private Double bestAttractFromPosition(Map<Integer, Position> takenTreasures, int initialDist, int deep) {
        
        Map<Position, Double> attract = new HashMap<Position, Double>(); 
        double actualAttract = 0;
        double maxAttract = 0;
        int robotDist, enemyDist, treasureValue;        
        
        Position actTreasure = takenTreasures.get(deep);
        
        Position actTreasurePosition = new Position(actTreasure);   
                
        if (tarjanWithEnergyRobot.distance(new Position(actTreasure.getX(), actTreasure.getY(), Robot.Direction.UP)) == tarjanWithEnergyRobot.distance(actTreasure)) 
            actTreasurePosition.setD(Robot.Direction.UP);                   
        else if (tarjanWithEnergyRobot.distance(new Position(actTreasure.getX(), actTreasure.getY(), Robot.Direction.RIGHT)) == tarjanWithEnergyRobot.distance(actTreasure)) 
            actTreasurePosition.setD(Robot.Direction.RIGHT); 
        else if (tarjanWithEnergyRobot.distance(new Position(actTreasure.getX(), actTreasure.getY(), Robot.Direction.DOWN)) == tarjanWithEnergyRobot.distance(actTreasure)) 
            actTreasurePosition.setD(Robot.Direction.DOWN); 
        else 
            actTreasurePosition.setD(Robot.Direction.LEFT); 
        
        for (Position treasure : treasures) {
            if (takenTreasures.containsValue(treasure)) {/*System.out.println("X");*/continue;} //musim vyhodit z hodnoceni ten poklad, kterej v tu dobu uz vlastne seberu 
            
            robotDist = initialDist + distance(actTreasurePosition, treasure);
            
            treasureValue = world.getTreasureValue(treasure.getX(), treasure.getY()); 
            
            actualAttract = (double) treasureValue / (robotDist + 1);
            
            for (Robot enemy : tarjanWithEnergyEnemies.keySet()) {
                enemyDist = deep+tarjanWithEnergyEnemies.get(enemy).distance(treasure);              
                if (robotDist > enemyDist) {
                    actualAttract *=  Math.sqrt((double) enemyDist / robotDist);                    
                }
            }
            
            if (deep+2 < Parameters.TREASURE_DEEP && actualAttract*(Parameters.TREASURE_DEEP-deep-1) > maxAttract) { //OPTIMALIACE aby se nesazil zlepsit to co uz nebude nejlepsi
                Map<Integer, Position> newTakenTreasures = new HashMap<Integer, Position>();
                newTakenTreasures.putAll(takenTreasures);
                newTakenTreasures.put(deep+1,treasure);                
                actualAttract +=  bestAttractFromPosition(newTakenTreasures, robotDist+1, deep+1);
            }            
            maxAttract = Math.max(actualAttract, maxAttract);            
        }     
        
        return maxAttract; 
    }    
    
       
    
    private int distance(Position treasure, Position position) {
        int distance;
        if (tarjanTreasures.containsKey(treasure.toInt())) {
            distance = tarjanTreasures.get(treasure.toInt()).distance(position);            
        } else {
            Tarjan treasureTarjan = new Tarjan(world);
            treasureTarjan.compute(treasure);
            tarjanTreasures.put(treasure.toInt(),treasureTarjan);      
            distance = treasureTarjan.distance(position);
        }      
        if (distance > world.getRoundsCount() - world.getRound())
            distance = Tarjan.NO_PATH;
        return distance;
    }
    
      
}
